(* robot.ml *)

open Unix

(********************)
(* Type declaration *)
(********************)

type tile = Space | Water | Wall | Home
type package = {idp:int;dnx:int;dny:int;wht:int}
type state = {mid:int;mutable cap:int;mutable pnx:int;mutable pny:int;
              mutable mon:int;mutable lpk:package list}
type robot = {idn:int;mutable psx:int;mutable psy:int}
type action = Move of int | Pick of int | Drop of int | Set of int*int | Nop

(**********)
(* Parser *)
(**********)

let parse_tile = function
  | '.' -> Space
  | '#' -> Wall
  | '~' -> Water
  | '@' -> Home
  | _ -> assert false

let parse_row s =
  let l = ref [] in
  begin
    for i=0 to ((String.length s)-1) do
      l := !l@[parse_tile (String.get s i)]
    done;
    !l
  end

let gram = Grammar.create (Plexer.make ())
let dim = Grammar.Entry.create gram "dim"
let til = Grammar.Entry.create gram "til"
let row = Grammar.Entry.create gram "row"
let mrt = Grammar.Entry.create gram "mrt"
let pos = Grammar.Entry.create gram "pos"
let cnf = Grammar.Entry.create gram "cnf"
let pak = Grammar.Entry.create gram "pak"
let alp = Grammar.Entry.create gram "alp"
let act = Grammar.Entry.create gram "act"
let oas = Grammar.Entry.create gram "oas"
let ans = Grammar.Entry.create gram "ans"
EXTEND
  dim: [ [ col = INT; row = INT -> (int_of_string col,int_of_string row) ] ];
  til: [ [ "." -> Space
         | "~" -> Water
         | "#" -> Wall
         | "@" -> Home ] ];
  row: [ [ l = LIST1 til -> l ] ];
  mrt: [ [ ein = INT; ecp=INT; emn=INT ->
           {mid=int_of_string ein;
            cap=int_of_string ecp;
            pnx=0;
            pny=0;
            mon=int_of_string emn;
            lpk=[]} ] ];
  pos: [ [ "."; ein=INT; UIDENT "X"; epx=INT; UIDENT "Y"; epy=INT ->
           {idn=int_of_string ein;
            psx=(int_of_string epx)-1;
            psy=(int_of_string epy)-1} ] ];
  cnf: [ [ l = LIST1 pos -> l ] ];
  pak: [ [ eip=INT; edx=INT; edy=INT; ewt=INT ->
           {idp=int_of_string eip;
            dnx=(int_of_string edx)-1;
            dny=(int_of_string edy)-1;
            wht=int_of_string ewt} ] ];
  alp: [ [ l = LIST0 pak -> l ] ];
  act: [ [ UIDENT "N" -> Move 0
         | UIDENT "S" -> Move 1
         | UIDENT "E" -> Move 2
         | UIDENT "W" -> Move 3
         | UIDENT "P"; idn=INT -> Pick (int_of_string idn)
         | UIDENT "D"; idn=INT -> Drop (int_of_string idn)
         | UIDENT "X"; stx=INT; UIDENT "Y"; sty=INT ->
           Set ((int_of_string stx)-1,(int_of_string sty)-1) ] ];
  oas: [ [ "."; idn=INT; l = LIST0 act ->
           (int_of_string idn,if l=[] then [Nop] else l) ] ];
  ans: [ [ l = LIST1 oas -> l ] ];
END

let parse_dim s = Grammar.Entry.parse dim (Stream.of_string s)
(*let parse_row s = Grammar.Entry.parse row (Stream.of_string s)*)
let parse_mrt s = Grammar.Entry.parse mrt (Stream.of_string s)
let parse_cnf s = Grammar.Entry.parse cnf (Stream.of_string s)
let parse_alp s = Grammar.Entry.parse alp (Stream.of_string s)
let parse_ans s = Grammar.Entry.parse ans (Stream.of_string s)

let sub_sharp s = 
  try
  begin
    while true do
      String.set s (String.index s '#') '.'
    done;
    s
  end
  with Not_found -> s

(******************)
(* Initialization *)
(******************)

let rec init acc sta = function
  | [] -> assert false
  | e::tl ->
    if sta.mid = e.idn then
    begin
      sta.pnx <- e.psx;
      sta.pny <- e.psy;
      (sta,acc@tl)
    end
    else init (acc@[e]) sta tl

let base_list (xdm,ydm) tab =
  let l = ref [] in
  begin
    for i=0 to (xdm-1) do
      for j=0 to (ydm-1) do
        if tab.(i).(j)=Home then l := !l@[i,j]
      done
    done;
    !l
  end

let read_board ic =
  let (xdm,ydm) = parse_dim (input_line ic) in
  let tab = Array.make_matrix xdm ydm Space in
  begin
    for i=0 to (ydm-1) do
      let l = parse_row (input_line ic) in
      for j=0 to (xdm-1) do
        tab.(j).(i) <- List.nth l j
      done
    done;
    let blt = base_list (xdm,ydm) tab
    and sta = parse_mrt (input_line ic) in
    let cnf = parse_cnf (sub_sharp (input_line ic)) in
    let (nsta,ncnf) = init [] sta cnf in
    (xdm,ydm,tab,blt,nsta,ncnf)
  end

let initialization ic oc =
  begin
    output_string oc "Player\n";
    flush oc;
    let brd =  read_board ic in
    print_endline "Initialization...Done";
    flush Pervasives.stdout;
    brd
  end

let client_fun ic oc = initialization ic oc

(*************)
(* Game play *)
(*************)

let ask_package ic = parse_alp (input_line ic)

let ask_answer ic = 
  try 
    let t = input_line ic in
    begin
      print_endline ("Turn result: "^t);
      flush Pervasives.stdout;
      parse_ans (sub_sharp t)
    end
  with End_of_file ->
    begin
      print_endline "Game over";
      flush Pervasives.stdout;
      exit 0
    end

(**** Updates ****)

let rec find_my_state id = function
  | [] -> assert false
  | (idn,lac)::tl ->
    if id=idn then tl
    else (idn,lac)::(find_my_state id tl)

let update_position x y = function
  | 0 -> (x,y+1)
  | 1 -> (x,y-1)
  | 2 -> (x+1,y)
  | 3 -> (x-1,y)
  | _ -> assert false

let rec give_package id = function
  | [] -> assert false
  | e::tl ->
    if id = e.idp then e
    else give_package id tl

let rec extract_package id = function
  | [] -> assert false
  | e::tl ->
    if id = e.idp then tl
    else e::(extract_package id tl)

let update_mrt_act lav sta = function
  | Move n -> let (x,y)=update_position sta.pnx sta.pny n in
              sta.pnx <- x;sta.pny <- y
  | Pick n -> sta.lpk <- sta.lpk@[(give_package n lav)]
  | Drop n -> sta.lpk <- extract_package n sta.lpk
  | Set(x,y) -> sta.pnx <- x;sta.pny <- y
  | Nop -> ()

let update_mrt lav sta lac = List.iter (update_mrt_act lav sta) lac

let update_rob_act rob = function
  | Move n -> let (x,y)=update_position rob.psx rob.psy n in
              rob.psx <- x;rob.psy <- y
  | Pick n -> ()
  | Drop n -> ()
  | Set(x,y) -> rob.psx <- x;rob.psy <- y
  | Nop -> ()

let update_rob rob lac = List.iter (update_rob_act rob) lac

let rec update_all_rob ans = function
  | [] -> []
  | e::tl ->
    (try update_rob e (List.assoc e.idn ans);e::(update_all_rob ans tl)
     with Not_found -> update_all_rob ans tl)

let rec mem_id_rob id = function
  | [] -> false
  | e::tl ->
    if e.idn = id then true
    else mem_id_rob id tl

let rec find_new_robots lrb = function
  | [] -> lrb
  | (id,lac)::tl ->
    if not(mem_id_rob id lrb) then
      let nrb = {idn=id;psx=0;psy=0} in
      begin
        update_rob nrb lac;
        find_new_robots (lrb@[nrb]) tl
      end
    else find_new_robots lrb tl

let main_update_rob lrb ans = find_new_robots (update_all_rob ans lrb) ans

let update sta rob lav ans =
  begin
    update_mrt lav sta (List.assoc sta.mid ans);
    let nlb = main_update_rob rob (find_my_state sta.mid ans) in
    (sta,nlb)
  end

(**** Our tactic ****)

let move = function
  | 0 -> "Move N"
  | 1 -> "Move S"
  | 2 -> "Move E"
  | 3 -> "Move W"
  | _ -> assert false

let pack act =
  let rec pack_rec = function
  | [] -> ""
  | id::tl -> " "^(string_of_int id)^(pack_rec tl) in
  fun lpk -> act^(pack_rec lpk)

let action bid act oc =
  begin
    output_string oc ((string_of_int bid)^" "^act^"\n");
    flush oc
  end

(**** Naive algorithm for lightning division ****)

let good_destination (x,y) pkg =
  if (x=pkg.dnx) && (y=pkg.dny) then pkg.idp
  else failwith "Not yet"

let try_drop (xdm,ydm,tab,blt,sta,rob) lav oc =
  let l = List.fold_left
          (fun a e -> try a@[good_destination (sta.pnx,sta.pny) e]
                      with _ -> a) [] sta.lpk in
  if l = [] then failwith "Nothing to drop"
  else action 1 (pack "Drop" l) oc
  
let try_pick (xdm,ydm,tab,blt,sta,rob) lav oc =
  let rec lpk_rec cap = function
    | [] -> []
    | pkg::tl ->
      if pkg.wht<=cap then pkg.idp::(lpk_rec (cap-pkg.wht) tl)
      else lpk_rec cap tl in
  let l = lpk_rec sta.cap lav in
  if l = [] then failwith "Nothing to pick"
  action 1 (pack "Pick" l) oc

let ok_case (xdm,ydm) tab (x,y) =
  not(x<0 || x>=xdm || y<0 || y >=ydm || tab.(x).(y)=Water || tab.(x).(y)=Wall)

let verify_dir (xdm,ydm) tab (x,y) = function
  | 0 -> ok_case (xdm,ydm) tab (x,y+1)
  | 1 -> ok_case (xdm,ydm) tab (x,y-1)
  | 2 -> ok_case (xdm,ydm) tab (x+1,y)
  | 3 -> ok_case (xdm,ydm) tab (x-1,y)
  | _ -> assert false

let rec try_move (xdm,ydm,tab,blt,sta,rob) lav oc =
  let dir = Random.int 4 in
  if verify_dir (xdm,ydm) tab (sta.pnx,sta.pny) dir then
    action 1 (move dir) oc
  else try_move (xdm,ydm,tab,blt,sta,rob) lav oc

let nop = action 1 (pack "Pick" [])

let drinker_algo (xdm,ydm,tab,blt,sta,rob) lav oc =
  try try_drop (xdm,ydm,tab,blt,sta,rob) lav oc with _ ->
  try try_pick (xdm,ydm,tab,blt,sta,rob) lav oc with _ ->
  try_move (xdm,ydm,tab,blt,sta,rob) lav oc

let tactic (xdm,ydm,tab,blt,sta,rob) lav oc =
  drinker_algo (xdm,ydm,tab,blt,sta,rob) lav oc

(**** Main loop ****)

let cpu_warning act oc =
  let t0 = Sys.time () in
  act oc;
  let t1 = Sys.time () in
  if (t1-.t0)>1. then
  begin
    print_endline "Warning: more than 1-CPUs for this turn";
    flush Pervasives.stdout
  end

let rec main_loop (xdm,ydm,tab,blt,sta,rob) ic oc =
  begin
    let lav = ask_package ic in
    print_endline "Package Info...Done";
    flush Pervasives.stdout;
    cpu_warning (tactic (xdm,ydm,tab,blt,sta,rob) lav) oc;
    let ans = ask_answer ic in
    let (nsta,nrob) = update sta rob lav ans in
    main_loop (xdm,ydm,tab,blt,nsta,nrob) ic oc
  end

(****************)
(* Socket stuff *)
(****************)

let domain_of = function
  | ADDR_UNIX _ -> PF_UNIX
  | ADDR_INET _ -> PF_INET

let open_connection sockaddr =
  let domain = domain_of sockaddr in
  let sock = Unix.socket domain Unix.SOCK_STREAM 0 
  in try Unix.connect sock sockaddr ;
     (Unix.in_channel_of_descr sock , Unix.out_channel_of_descr sock)
     with exn -> Unix.close sock ; raise exn

let shutdown_connection inchan =
  Unix.shutdown (Unix.descr_of_in_channel inchan) Unix.SHUTDOWN_SEND ;;

let main_client client_fun =
  if Array.length Sys.argv < 3 
  then Printf.printf "Usage: runme serveur port\n"
  else let serveur = Sys.argv.(1) in
  let serveur_adr =
    try  Unix.inet_addr_of_string serveur 
    with Failure("inet_addr_of_string") -> 
      try  (Unix.gethostbyname serveur).Unix.h_addr_list.(0) 
      with Not_found ->
        Printf.eprintf "%s : serveur inconnu\n" serveur ;
        exit 2
  in try 
     let port = int_of_string (Sys.argv.(2)) in
     let sockadr = Unix.ADDR_INET(serveur_adr,port) in 
     let ic,oc = open_connection sockadr
     in client_fun ic oc ;
     shutdown_connection ic
     with Failure("int_of_string") -> Printf.eprintf "bad port number";
                                      exit 2

let main ic oc = main_loop (initialization ic oc) ic oc

let _ =
  begin
    Random.self_init ();
    main_client main
  end

(*let _ = initialization Pervasives.stdin Pervasives.stdout*)
